# This file is part of pipe_base.
#
# Developed for the LSST Data Management System.
# This product includes software developed by the LSST Project
# (http://www.lsst.org).
# See the COPYRIGHT file at the top-level directory of this distribution
# for details of code ownership.
#
# This software is dual licensed under the GNU General Public License and also
# under a 3-clause BSD license. Recipients may choose which of these licenses
# to use; please see the files gpl-3.0.txt and/or bsd_license.txt,
# respectively.  If you choose the GPL option then the following text applies
# (but note that there is still no warranty even if you opt for BSD instead):
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

from __future__ import annotations

__all__ = ("aggregate_graph",)

import dataclasses
import itertools
import uuid

import astropy.units as u
import networkx

from lsst.utils.logging import getLogger
from lsst.utils.usage import get_peak_mem_usage

from ...graph_walker import GraphWalker
from ...pipeline_graph import TaskImportMode
from .._predicted import PredictedQuantumGraphComponents, PredictedQuantumGraphReader
from .._provenance import ProvenanceQuantumScanData, ProvenanceQuantumScanStatus
from ._communicators import (
    IngesterCommunicator,
    ScannerCommunicator,
    SupervisorCommunicator,
    WriterCommunicator,
)
from ._config import AggregatorConfig
from ._ingester import Ingester
from ._scanner import Scanner
from ._structs import ScanReport
from ._workers import SpawnWorkerFactory, ThreadWorkerFactory
from ._writer import Writer


@dataclasses.dataclass
class Supervisor:
    """The main process/thread for the provenance aggregator."""

    predicted_path: str
    """Path to the predicted quantum graph."""

    comms: SupervisorCommunicator
    """Communicator object for the supervisor."""

    predicted: PredictedQuantumGraphComponents = dataclasses.field(init=False)
    """Components of the predicted quantum graph."""

    walker: GraphWalker[uuid.UUID] = dataclasses.field(init=False)
    """Iterator that traverses the quantum graph."""

    n_abandoned: int = 0
    """Number of quanta we abandoned because they did not complete in time and
    we could not assume they had failed.
    """

    def __post_init__(self) -> None:
        self.comms.progress.log.info("Reading predicted quantum graph.")
        with PredictedQuantumGraphReader.open(
            self.predicted_path, import_mode=TaskImportMode.DO_NOT_IMPORT
        ) as reader:
            reader.read_thin_graph()
            reader.read_init_quanta()
            self.predicted = reader.components
        self.comms.progress.log.info("Analyzing predicted graph.")
        xgraph = networkx.DiGraph(self.predicted.thin_graph.edges)
        # Make sure all quanta are in the graph, even if they don't have any
        # quantum-only edges.
        for thin_quantum in itertools.chain.from_iterable(self.predicted.thin_graph.quanta.values()):
            xgraph.add_node(thin_quantum.quantum_id)
        # Add init quanta as nodes without edges, because the scanner should
        # only be run after init outputs are all written and hence we don't
        # care when we process them.
        for init_quantum in self.predicted.init_quanta.root:
            xgraph.add_node(init_quantum.quantum_id)
        self.walker = GraphWalker(xgraph)

    def loop(self) -> None:
        """Scan the outputs of the quantum graph to gather provenance and
        ingest outputs.
        """
        n_quanta = self.predicted.header.n_quanta + len(self.predicted.init_quanta.root)
        self.comms.progress.scans.total = n_quanta
        self.comms.progress.writes.total = n_quanta
        self.comms.progress.quantum_ingests.total = n_quanta
        ready_set: set[uuid.UUID] = set()
        for ready_quanta in self.walker:
            self.comms.log.debug("Sending %d new quanta to scan queue.", len(ready_quanta))
            ready_set.update(ready_quanta)
            while ready_set:
                self.comms.request_scan(ready_set.pop())
            for scan_return in self.comms.poll():
                self.handle_report(scan_return)
        if self.comms.config.incomplete:
            quantum_or_quanta = "quanta" if self.n_abandoned != 1 else "quantum"
            self.comms.progress.log.info(
                "%d %s incomplete/failed abandoned; re-run with incomplete=False to finish.",
                self.n_abandoned,
                quantum_or_quanta,
            )
        self.comms.progress.log.info(
            "Scanning complete after %0.1fs; waiting for workers to finish.",
            self.comms.progress.elapsed_time,
        )

    def handle_report(self, scan_report: ScanReport) -> None:
        """Handle a report from a scanner.

        Parameters
        ----------
        scan_report : `ScanReport`
            Information about the scan.
        """
        match scan_report.status:
            case ProvenanceQuantumScanStatus.SUCCESSFUL | ProvenanceQuantumScanStatus.INIT:
                self.comms.log.debug("Scan complete for %s: quantum succeeded.", scan_report.quantum_id)
                self.walker.finish(scan_report.quantum_id)
            case ProvenanceQuantumScanStatus.FAILED:
                self.comms.log.debug("Scan complete for %s: quantum failed.", scan_report.quantum_id)
                blocked_quanta = self.walker.fail(scan_report.quantum_id)
                for blocked_quantum_id in blocked_quanta:
                    if self.comms.config.is_writing_provenance:
                        self.comms.request_write(
                            ProvenanceQuantumScanData(
                                blocked_quantum_id, status=ProvenanceQuantumScanStatus.BLOCKED
                            )
                        )
                    self.comms.progress.scans.update(1)
                self.comms.progress.quantum_ingests.update(len(blocked_quanta))
            case ProvenanceQuantumScanStatus.ABANDONED:
                self.comms.log.debug("Abandoning scan for %s: quantum has not succeeded (yet).")
                self.walker.fail(scan_report.quantum_id)
                self.n_abandoned += 1
            case unexpected:
                raise AssertionError(
                    f"Unexpected status {unexpected!r} in scanner loop for {scan_report.quantum_id}."
                )
        self.comms.progress.scans.update(1)


def aggregate_graph(predicted_path: str, butler_path: str, config: AggregatorConfig) -> None:
    """Run the graph aggregator tool.

    Parameters
    ----------
    predicted_path : `str`
        Path to the predicted quantum graph.
    butler_path : `str`
        Path or alias to the central butler repository.
    config : `AggregatorConfig`
        Configuration for the aggregator.
    """
    log = getLogger("lsst.pipe.base.quantum_graph.aggregator")
    worker_factory = ThreadWorkerFactory() if config.n_processes == 1 else SpawnWorkerFactory()
    with SupervisorCommunicator(log, config.n_processes, worker_factory, config) as comms:
        comms.progress.log.verbose("Starting workers.")
        if config.is_writing_provenance:
            writer_comms = WriterCommunicator(comms)
            comms.workers[writer_comms.name] = worker_factory.make_worker(
                target=Writer.run,
                args=(predicted_path, writer_comms),
                name=writer_comms.name,
            )
        for scanner_id in range(config.n_processes):
            scanner_comms = ScannerCommunicator(comms, scanner_id)
            comms.workers[scanner_comms.name] = worker_factory.make_worker(
                target=Scanner.run,
                args=(predicted_path, butler_path, scanner_comms),
                name=scanner_comms.name,
            )
        ingester_comms = IngesterCommunicator(comms)
        comms.workers[ingester_comms.name] = worker_factory.make_worker(
            target=Ingester.run,
            args=(predicted_path, butler_path, ingester_comms),
            name=ingester_comms.name,
        )
        supervisor = Supervisor(predicted_path, comms)
        supervisor.loop()
    # We can't get memory usage for children until they've joined.
    parent_mem, child_mem = get_peak_mem_usage()
    # This is actually an upper bound on the peak (since the peaks could be
    # at different times), but since we expect memory usage to be more smooth
    # than spiky that's fine.
    total_mem: u.Quantity = parent_mem + child_mem
    log.info(
        "All aggregation tasks complete after %0.1fs; peak memory usage ≤ %0.1f MB.",
        comms.progress.elapsed_time,
        total_mem.to(u.MB).value,
    )



# This file is part of pipe_base.
#
# Developed for the LSST Data Management System.
# This product includes software developed by the LSST Project
# (https://www.lsst.org).
# See the COPYRIGHT file at the top-level directory of this distribution
# for details of code ownership.
#
# This software is dual licensed under the GNU General Public License and also
# under a 3-clause BSD license. Recipients may choose which of these licenses
# to use; please see the files gpl-3.0.txt and/or bsd_license.txt,
# respectively.  If you choose the GPL option then the following text applies
# (but note that there is still no warranty even if you opt for BSD instead):
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <https://www.gnu.org/licenses/>.

import itertools
import logging
import os
import shutil
import subprocess
import tempfile
import unittest
import uuid
from io import StringIO

import lsst.utils.tests
from lsst.pipe.base import QuantumGraph
from lsst.pipe.base.dot_tools import graph2dot
from lsst.pipe.base.mermaid_tools import graph2mermaid
from lsst.pipe.base.quantum_graph import (
    FORMAT_VERSION,
    AddressReader,
    PredictedDatasetInfo,
    PredictedQuantumGraph,
    PredictedQuantumInfo,
)
from lsst.pipe.base.tests.mocks import DynamicConnectionConfig, InMemoryRepo

logging.getLogger("timer").setLevel(logging.INFO)
logging.getLogger("lsst").setLevel(logging.INFO)
logging.getLogger("lsst.pipe.base.quantum_graph").setLevel(logging.DEBUG)


class PredictedQuantumGraphTestCase(unittest.TestCase):
    """Unit tests for the predicted quantum graph classes.

    These tests focus on the interface of the predicted quantum graph
    especially I/O and conversion to and from the old quantum graph class.

    Additional test coverage is provided by the tests for quantum graph
    builders and executors, which all use this class as well.
    """

    def setUp(self):
        self.helper = InMemoryRepo("base.yaml", "spatial.yaml")
        self.enterContext(self.helper)
        self.helper.add_task(
            "calibrate",
            dimensions=["visit", "detector"],
            inputs={
                "input_image": DynamicConnectionConfig(
                    dataset_type_name="raw",
                    dimensions=["visit", "detector"],
                )
            },
            prerequisite_inputs={
                "refcat": DynamicConnectionConfig(
                    dataset_type_name="references",
                    dimensions=["htm7"],
                    multiple=True,
                )
            },
            init_outputs={
                "output_schema": DynamicConnectionConfig(
                    dataset_type_name="source_schema",
                )
            },
            outputs={
                "output_image": DynamicConnectionConfig(
                    dataset_type_name="image",
                    dimensions=["visit", "detector"],
                ),
                "output_table": DynamicConnectionConfig(
                    dataset_type_name="source_detector",
                    dimensions=["visit", "detector"],
                ),
            },
        )
        self.helper.add_task(
            "consolidate",
            dimensions=["visit"],
            init_inputs={
                "input_schema": DynamicConnectionConfig(
                    dataset_type_name="source_schema",
                )
            },
            inputs={
                "input_table": DynamicConnectionConfig(
                    dataset_type_name="source_detector",
                    dimensions=["visit", "detector"],
                    multiple=True,
                )
            },
            outputs={
                "output_table": DynamicConnectionConfig(
                    dataset_type_name="source",
                    dimensions=["visit"],
                )
            },
        )
        self.helper.add_task(
            "resample",
            dimensions=["patch", "visit"],
            inputs={
                "input_image": DynamicConnectionConfig(
                    dataset_type_name="image",
                    dimensions=["visit", "detector"],
                    multiple=True,
                )
            },
            outputs={
                "output_image": DynamicConnectionConfig(
                    dataset_type_name="warp",
                    dimensions=["patch", "visit"],
                )
            },
        )
        self.helper.add_task(
            "coadd",
            dimensions=["patch", "band"],
            inputs={
                "input_image": DynamicConnectionConfig(
                    dataset_type_name="warp",
                    dimensions=["patch", "visit"],
                    multiple=True,
                )
            },
            outputs={
                "output_image": DynamicConnectionConfig(
                    dataset_type_name="coadd",
                    dimensions=["patch", "band"],
                ),
                "output_table": DynamicConnectionConfig(
                    dataset_type_name="object",
                    dimensions=["patch", "band"],
                    # Like all other (defaulted) storage classes here,
                    # 'ArrowAstropy' below will be mocked; we pick it so we can
                    # try a component input, and because we know its pytype is
                    # safe to import here.
                    storage_class="ArrowAstropy",
                ),
            },
        )
        self.helper.add_task(
            "match",
            dimensions=["htm6"],
            inputs={
                "input_object": DynamicConnectionConfig(
                    dataset_type_name="object",
                    dimensions=["patch", "band"],
                    multiple=True,
                    storage_class="ArrowAstropy",
                ),
                "input_source": DynamicConnectionConfig(
                    dataset_type_name="source",
                    dimensions=["visit"],
                    multiple=True,
                ),
                "input_object_schema": DynamicConnectionConfig(
                    dataset_type_name="object.schema",
                    dimensions=["visit"],
                    multiple=True,
                    storage_class="ArrowAstropySchema",
                ),
            },
            outputs={
                "output_table": DynamicConnectionConfig(
                    dataset_type_name="matches",
                    dimensions=["htm6"],
                )
            },
        )
        self.builder = self.helper.make_quantum_graph_builder()

    def check_quantum_graph(
        self,
        qg: PredictedQuantumGraph,
        *,
        all_tasks: bool = True,
        thin_graph: bool = True,
        all_quantum_datasets: bool = True,
        init_quanta: bool = True,
        dimension_data_loaded: bool = True,
        dimension_data_deserialized: bool = True,
        converting_partial: bool = False,
    ) -> None:
        """Test the attributes and methods of a quantum graph produced by this
        test case's builder.

        Parameters
        ----------
        qg : `lsst.pipe.base.quantum_graph.PredictedQuantumGraph`
            Graph to test.
        all_tasks : `bool`, optional
            Whether all tasks in the pipeline were loaded.
        thin_graph : `bool`, optional
            Whether the ``thin_graph`` component was loaded.
        all_quantum_datasets : `bool`, optional
            Whether all quantum datasets were loaded.
        init_quanta : `bool`, optional
            Whether the ``init_quanta`` component was loaded.
        dimension_data_loaded : `bool`, optional
            Whether the ``dimension_data`` component was loaded at all.
        dimension_data_deserialized : `bool`, optional
            Whether all dimension records are expected to have been
            deserialized.  Ignored if ``dimension_data_loaded`` is `False`.
        converting_partial : `bool`, optional
            Whether this was a partial load that only included some quanta.
        """
        self.assertEqual(qg.header.inputs, ["input_chain"])
        self.assertEqual(qg.header.output, "output_chain")
        self.assertEqual(qg.header.output_run, "output_run")
        self.assertEqual(qg.header.metadata["stuff"], "whatever")
        self.assertEqual(qg.header.version, FORMAT_VERSION)
        if not converting_partial:
            # When partial-reading an old QG, counts reflect what was loaded.
            # In all other cases they reflect the full original graph.
            self.assertEqual(qg.header.n_task_quanta["calibrate"], 8)
            self.assertEqual(qg.header.n_task_quanta["consolidate"], 2)
            self.assertEqual(qg.header.n_task_quanta["resample"], 10)
            self.assertEqual(qg.header.n_task_quanta["coadd"], 10)
            self.assertEqual(qg.header.n_task_quanta["match"], 2)
            self.assertEqual(qg.header.n_quanta, 32)
            self.assertEqual(
                qg.header.n_datasets,
                63  # non-automatic datasets (see breakdown below)
                + 32 * 2  # one metadata and one log for each quantam
                + 5  # one config for each task
                + 1,  # packages
            )
        if all_tasks:
            self.assertFalse(self.helper.pipeline_graph.diff_tasks(qg.pipeline_graph))
        if thin_graph or all_quantum_datasets:
            self.assertEqual(len(qg), qg.header.n_quanta)
            self.assertEqual(len(qg.quantum_only_xgraph), qg.header.n_quanta)
            self.assertEqual(len(qg.quanta_by_task["calibrate"]), 8)
            self.assertEqual(len(qg.quanta_by_task["consolidate"]), 2)
            self.assertEqual(len(qg.quanta_by_task["resample"]), 10)
            self.assertEqual(len(qg.quanta_by_task["coadd"]), 10)
            self.assertEqual(len(qg.quanta_by_task["match"]), 2)
        if init_quanta:
            if all_tasks:
                self.assertEqual(len(qg.datasets_by_type["calibrate_config"]), 1)
                self.assertEqual(len(qg.datasets_by_type["consolidate_config"]), 1)
                self.assertEqual(len(qg.datasets_by_type["resample_config"]), 1)
                self.assertEqual(len(qg.datasets_by_type["coadd_config"]), 1)
                self.assertEqual(len(qg.datasets_by_type["match_config"]), 1)
                self.assertEqual(len(qg.datasets_by_type["source_schema"]), 1)
            self.assertEqual(len(qg.datasets_by_type["packages"]), 1)
            n_init_datasets = 7
        else:
            n_init_datasets = 0
        if all_quantum_datasets:
            assert all_tasks
            self.assertEqual(
                len(qg.bipartite_xgraph),
                qg.header.n_quanta + qg.header.n_datasets - n_init_datasets,
            )
            self.assertEqual(len(qg.datasets_by_type["raw"]), 8)
            self.assertEqual(len(qg.datasets_by_type["references"]), 4)
            self.assertEqual(len(qg.datasets_by_type["image"]), 8)
            self.assertEqual(len(qg.datasets_by_type["source_detector"]), 8)
            self.assertEqual(len(qg.datasets_by_type["source"]), 2)
            self.assertEqual(len(qg.datasets_by_type["warp"]), 10)
            self.assertEqual(len(qg.datasets_by_type["coadd"]), 10)
            self.assertEqual(len(qg.datasets_by_type["object"]), 10)
            self.assertEqual(len(qg.datasets_by_type["matches"]), 2)
            # Spot-check some edges and graph structure.
            for data_id, dataset_id in qg.datasets_by_type["source_detector"].items():
                for producer_id, _, edge_data in qg.bipartite_xgraph.in_edges(dataset_id, data=True):
                    self.assertIn(producer_id, qg.quanta_by_task["calibrate"].values())
                    self.assertFalse(edge_data["is_read"])
                    self.assertEqual(
                        edge_data["pipeline_edges"],
                        [qg.pipeline_graph.tasks["calibrate"].get_output_edge("output_table")],
                    )
                for _, consumer_id, edge_data in qg.bipartite_xgraph.out_edges(dataset_id, data=True):
                    self.assertIn(consumer_id, qg.quanta_by_task["consolidate"].values())
                    self.assertTrue(edge_data["is_read"])
                    self.assertEqual(
                        edge_data["pipeline_edges"],
                        [qg.pipeline_graph.tasks["consolidate"].get_input_edge("input_table")],
                    )
        # We use 'is' checks instead of just equality because we don't want
        # duplicates of these objects floating around wasting memory.
        for task_label, quanta_for_task in qg.quanta_by_task.items():
            for data_id, quantum_id in quanta_for_task.items():
                d1: PredictedQuantumInfo = qg.quantum_only_xgraph.nodes[quantum_id]
                self.assertEqual(d1["task_label"], task_label)
                self.assertIs(d1["data_id"], data_id)
                self.assertIs(d1["pipeline_node"], qg.pipeline_graph.tasks[task_label])
                d2: PredictedQuantumInfo = qg.bipartite_xgraph.nodes[quantum_id]
                self.assertEqual(d2["task_label"], task_label)
                self.assertIs(d2["data_id"], data_id)
                self.assertIs(d2["pipeline_node"], qg.pipeline_graph.tasks[task_label])
        for dataset_type_name, datasets_for_type in qg.datasets_by_type.items():
            if (
                dataset_type_name == "source_schema"
                or dataset_type_name == "packages"
                or dataset_type_name.endswith("_config")
            ):
                continue
            for data_id, dataset_id in datasets_for_type.items():
                d3: PredictedDatasetInfo = qg.bipartite_xgraph.nodes[dataset_id]
                self.assertIs(d3["data_id"], data_id)
                self.assertIs(d3["pipeline_node"], qg.pipeline_graph.dataset_types[dataset_type_name])
        if dimension_data_loaded:
            self.assertIsNotNone(qg.dimension_data)
            if dimension_data_deserialized and not converting_partial:
                self.assertEqual(len(qg.dimension_data.records["instrument"]), 1)
                self.assertEqual(len(qg.dimension_data.records["visit"]), 2)
                self.assertEqual(len(qg.dimension_data.records["detector"]), 4)
                self.assertEqual(len(qg.dimension_data.records["visit_detector_region"]), 8)
                self.assertEqual(len(qg.dimension_data.records["tract"]), 2)
                self.assertEqual(len(qg.dimension_data.records["patch"]), 9)
                self.assertEqual(len(qg.dimension_data.records["band"]), 2)
                self.assertEqual(len(qg.dimension_data.records["physical_filter"]), 2)

    def define_partial_read(self, qg: PredictedQuantumGraph) -> list[uuid.UUID]:
        """Return a list of quantum IDs to test partial loads.

        Parameters
        ----------
        qg : `lsst.pipe.base.quantum_graph.PredictedQuantumGraph`
            Full graph to obtain quantum IDs from.  Note that quantum IDs are
            random and change every test run, and hence test behavior should
            never depend on explicit values or their sort order.

        Returns
        -------
        quanta : `list` [ `uuid.UUID` ]
            Quanta to use for partial-read tests.  These always have the same
            task labels and data IDs: all ``calibrate`` and ``consolidate``
            quanta for ``visit==1``.
        """
        return [
            quantum_id
            for data_id, quantum_id in qg.quanta_by_task["calibrate"].items()
            if data_id["visit"] == 1
        ] + [
            quantum_id
            for data_id, quantum_id in qg.quanta_by_task["consolidate"].items()
            if data_id["visit"] == 1
        ]

    def check_partial_read(
        self, qg: PredictedQuantumGraph, quanta: list[uuid.UUID], converting: bool = False
    ) -> None:
        """Test the attributes and methods of a partially-loaded quantum graph.

        Parameters
        ----------
        qg : `lsst.pipe.base.quantum_graph.PredictedQuantumGraph`
            Graph for which only ``quanta`` were loaded.
        quanta : `list` [ `uuid.UUID` ]
            IDs of the quanta that were loaded.  Must be the result of a call
            to `define_partial_read`.
        converting : `bool`, optional
            Whether this graph was read via the old
            `lsst.pipe.base.QuantumGraph` class and then converted.
        """
        self.check_quantum_graph(
            qg,
            all_tasks=False,
            thin_graph=False,
            all_quantum_datasets=False,
            converting_partial=converting,
            dimension_data_deserialized=converting,
        )
        self.assertEqual(len(qg), 5)
        self.assertEqual(len(qg.quantum_only_xgraph), 5)
        execution_quanta = qg.build_execution_quanta()
        self.assertEqual(execution_quanta.keys(), set(quanta))
        for quantum_id, quantum in execution_quanta.items():
            self.assertIs(qg.bipartite_xgraph.nodes[quantum_id]["quantum"], quantum)
            self.assertIs(qg.quantum_only_xgraph.nodes[quantum_id]["quantum"], quantum)

    def test_build(self) -> None:
        """Test building a `PredictedQuantumGraph` by
        inspecting the result.
        """
        qg = self.builder.finish(
            output="output_chain",
            metadata={"stuff": "whatever"},
            attach_datastore_records=False,
        ).assemble()
        self.check_quantum_graph(qg)

    def test_from_old_quantum_graph(self) -> None:
        """Test building a old `QuantumGraph` and then
        converting it to a `PredictedQuantumGraph`.
        """
        old_qg = self.builder.build(
            metadata={
                "input": ["input_chain"],
                "output": "output_chain",
                "output_run": "output_run",
                "stuff": "whatever",
            },
            attach_datastore_records=False,
        )
        new_qg = PredictedQuantumGraph.from_old_quantum_graph(old_qg)
        self.check_quantum_graph(new_qg)

    def test_read_execution_quanta_old_file(self) -> None:
        """Test building a old `QuantumGraph`, saving it, and then reading it
        via `PredictedQuantumGraph.read_execution_quanta`.
        """
        old_qg = self.builder.build(
            metadata={
                "input": ["input_chain"],
                "output": "output_chain",
                "output_run": "output_run",
                "stuff": "whatever",
            },
            attach_datastore_records=False,
        )
        with tempfile.TemporaryDirectory(ignore_cleanup_errors=True) as tmpdir:
            tmpfile = os.path.join(tmpdir, "old.qgraph")
            old_qg.saveUri(tmpfile)
            # Read everything.
            qg1 = PredictedQuantumGraph.read_execution_quanta(tmpfile)
            self.check_quantum_graph(qg1, thin_graph=False)
            # Read just a few quanta.
            quanta = self.define_partial_read(qg1)
            qg2 = PredictedQuantumGraph.read_execution_quanta(tmpfile, quantum_ids=quanta)
            self.check_partial_read(qg2, quanta, converting=True)

    def test_roundtrip_old_quantum_graph(self) -> None:
        """Test building a `PredictedQuantumGraph` and round-tripping it
        through the old `QuantumGraph` class.
        """
        qg1 = self.builder.finish(
            output="output_chain",
            metadata={"stuff": "whatever"},
            attach_datastore_records=False,
        ).assemble()
        old_qg = qg1.to_old_quantum_graph()
        qg2 = PredictedQuantumGraph.from_old_quantum_graph(old_qg)
        self.check_quantum_graph(qg2)

    def test_write_new_as_old(self) -> None:
        """Test building a `PredictedQuantumGraphComponents`, saving it with
        the old extension, and reading it back in both the old class and the
        new class.
        """
        components = self.builder.finish(
            output="output_chain",
            metadata={"stuff": "whatever"},
            attach_datastore_records=False,
        )
        with tempfile.TemporaryDirectory(ignore_cleanup_errors=True) as tmpdir:
            tmpfile = os.path.join(tmpdir, "old.qgraph")
            components.write(tmpfile)
            qg1 = PredictedQuantumGraph.read_execution_quanta(tmpfile)
            self.check_quantum_graph(qg1)
            old_qg = QuantumGraph.loadUri(tmpfile)
            qg2 = PredictedQuantumGraph.from_old_quantum_graph(old_qg)
            self.check_quantum_graph(qg2)

    def test_read_new_as_old(self) -> None:
        """Test building a `PredictedQuantumGraphComponents`, saving it with
        the new extension, and reading it back in via the old class.
        """
        components = self.builder.finish(
            output="output_chain",
            metadata={"stuff": "whatever"},
            attach_datastore_records=False,
        )
        with tempfile.TemporaryDirectory(ignore_cleanup_errors=True) as tmpdir:
            tmpfile = os.path.join(tmpdir, "new.qg")
            components.write(tmpfile)
            old_qg = QuantumGraph.loadUri(tmpfile)
            new_qg = PredictedQuantumGraph.from_old_quantum_graph(old_qg)
            self.check_quantum_graph(new_qg)

    def test_io(self) -> None:
        """Test saving a `PredictedQuantumGraphComponents` and reading it back
        in as a `PredictedQuantumGraph`, both fully and partially, and as a
        `QuantumGraph`, both fully and partially.
        """
        components = self.builder.finish(
            output="output_chain",
            metadata={"stuff": "whatever"},
            attach_datastore_records=False,
        )
        # We use a small page size since this is a tiny graph and hence
        # otherwise all addresses would fit in a single page.
        four_row_page_size = AddressReader.compute_row_size(int_size=8, n_addresses=1) * 4
        with tempfile.TemporaryDirectory(ignore_cleanup_errors=True) as tmpdir:
            tmpfile = os.path.join(tmpdir, "new.qg")
            components.write(tmpfile, zstd_dict_n_inputs=24)  # enable dict compression code path
            # Test a full read with the new class.
            with PredictedQuantumGraph.open(tmpfile, page_size=four_row_page_size) as reader:
                reader.read_all()
                full_qg = reader.finish()
                self.check_quantum_graph(full_qg, dimension_data_deserialized=False)
            # Test a full read with the old class (uses new class and then
            # converts to old, and we convert back to new for the test).
            self.check_quantum_graph(
                PredictedQuantumGraph.from_old_quantum_graph(QuantumGraph.loadUri(tmpfile))
            )
            # Test a partial read with the old class.
            quanta = self.define_partial_read(full_qg)
            self.check_partial_read(
                PredictedQuantumGraph.from_old_quantum_graph(QuantumGraph.loadUri(tmpfile, nodes=quanta)),
                quanta,
                converting=True,
            )
            # Test a thin but shallow read with the new class.
            with PredictedQuantumGraph.open(tmpfile, page_size=four_row_page_size) as reader:
                reader.read_thin_graph()
                thin_qg = reader.finish()
                self.check_quantum_graph(
                    thin_qg,
                    dimension_data_deserialized=False,
                    all_quantum_datasets=False,
                    dimension_data_loaded=False,
                    init_quanta=False,
                )
            # Test a deep read of just a few quanta with the new class.
            narrow_qg = PredictedQuantumGraph.read_execution_quanta(
                tmpfile,
                quantum_ids=quanta,
                page_size=four_row_page_size,
            )
            self.check_partial_read(narrow_qg, quanta)

    def test_no_compression_dict(self) -> None:
        """Test saving a `PredictedQuantumGraphComponents` and reading it back
        in as a `PredictedQuantumGraph` with the compression dictionary
        disabled.
        """
        components = self.builder.finish(
            output="output_chain",
            metadata={"stuff": "whatever"},
            attach_datastore_records=False,
        )
        # We use a small page size since this is a tiny graph and hence
        # otherwise all addresses would fit in a single page.
        three_row_page_size = AddressReader.compute_row_size(int_size=8, n_addresses=1) * 3
        with tempfile.TemporaryDirectory(ignore_cleanup_errors=True) as tmpdir:
            tmpfile = os.path.join(tmpdir, "new.qg")
            components.write(tmpfile, zstd_dict_size=0)
            with PredictedQuantumGraph.open(tmpfile, page_size=three_row_page_size) as reader:
                reader.read_all()
                full_qg = reader.finish()
                self.check_quantum_graph(full_qg, dimension_data_deserialized=False)

    def test_dot(self) -> None:
        """Test visualization via GraphViz dot."""
        qg = self.builder.finish(attach_datastore_records=False).assemble()
        stream = StringIO()
        graph2dot(qg, stream)
        if (dot_cmd := shutil.which("dot")) is None:
            raise unittest.SkipTest("Aborting test early; 'dot' command is not available.")
        # Just check that the dot command can parse what we've given it.
        result = subprocess.run(
            [dot_cmd, "-Txdot"], input=stream.getvalue().encode(), stdout=subprocess.DEVNULL
        )
        result.check_returncode()

    def test_mermaid(self) -> None:
        """Test visualization via Mermaid."""
        qg = self.builder.finish(attach_datastore_records=False).assemble()
        stream = StringIO()
        # Just check that it runs without error.
        graph2mermaid(qg, stream)

    def test_update_output_run(self) -> None:
        """Test the PredictedQuantumGraphComponents.output_run method."""
        components = self.builder.finish(
            output="output_chain",
            metadata={"stuff": "whatever"},
            attach_datastore_records=False,
        )
        components.update_output_run("new_output_run")
        qg = components.assemble()
        for ref in qg.get_init_outputs("calibrate").values():
            self.assertEqual(ref.run, "new_output_run")
        for quantum in qg.build_execution_quanta().values():
            for ref in itertools.chain.from_iterable(quantum.outputs.values()):
                self.assertEqual(ref.run, "new_output_run", msg=str(ref))
            for ref in itertools.chain.from_iterable(quantum.inputs.values()):
                if ref.datasetType.name not in ("raw", "references"):
                    self.assertEqual(ref.run, "new_output_run", msg=str(ref))


if __name__ == "__main__":
    lsst.utils.tests.init()
    unittest.main()



# This file is part of pipe_base.
#
# Developed for the LSST Data Management System.
# This product includes software developed by the LSST Project
# (http://www.lsst.org).
# See the COPYRIGHT file at the top-level directory of this distribution
# for details of code ownership.
#
# This software is dual licensed under the GNU General Public License and also
# under a 3-clause BSD license. Recipients may choose which of these licenses
# to use; please see the files gpl-3.0.txt and/or bsd_license.txt,
# respectively.  If you choose the GPL option then the following text applies
# (but note that there is still no warranty even if you opt for BSD instead):
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

from __future__ import annotations

__all__ = ("IngestRequest", "ScanReport")

import dataclasses
import uuid

from lsst.daf.butler import DatasetRef
from lsst.daf.butler.datastore.record_data import DatastoreRecordData

from .._common import DatastoreName
from .._provenance import ProvenanceQuantumScanStatus


@dataclasses.dataclass
class ScanReport:
    """Minimal information needed about a completed scan by the supervisor."""

    quantum_id: uuid.UUID
    """Unique ID of the quantum."""

    status: ProvenanceQuantumScanStatus
    """Combined status of the scan and the execution of the quantum."""


@dataclasses.dataclass
class IngestRequest:
    """A request to ingest datasets produced by a single quantum."""

    producer_id: uuid.UUID
    """ID of the quantum that produced these datasets."""

    refs: list[DatasetRef]
    """Registry information about the datasets."""

    records: dict[DatastoreName, DatastoreRecordData]
    """Datastore information about the datasets."""

    def __bool__(self) -> bool:
        return bool(self.refs or self.records)



# This file is part of pipe_base.
#
# Developed for the LSST Data Management System.
# This product includes software developed by the LSST Project
# (http://www.lsst.org).
# See the COPYRIGHT file at the top-level directory of this distribution
# for details of code ownership.
#
# This software is dual licensed under the GNU General Public License and also
# under a 3-clause BSD license. Recipients may choose which of these licenses
# to use; please see the files gpl-3.0.txt and/or bsd_license.txt,
# respectively.  If you choose the GPL option then the following text applies
# (but note that there is still no warranty even if you opt for BSD instead):
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

from __future__ import annotations

__all__ = (
    "FatalWorkerError",
    "IngesterCommunicator",
    "ScannerCommunicator",
    "SupervisorCommunicator",
)

import cProfile
import dataclasses
import enum
import logging
import os
import signal
import time
import uuid
from collections.abc import Iterable, Iterator
from contextlib import ExitStack
from traceback import format_exception
from types import TracebackType
from typing import Literal, Self, overload

from lsst.utils.logging import LsstLogAdapter

from .._provenance import ProvenanceQuantumScanData
from ._config import AggregatorConfig
from ._progress import ProgressManager, make_worker_log
from ._structs import IngestRequest, ScanReport
from ._workers import Event, Queue, Worker, WorkerFactory

_TINY_TIMEOUT = 0.01


class FatalWorkerError(BaseException):
    """An exception raised by communicators when one worker (including the
    supervisor) has caught an exception in order to signal the others to shut
    down.
    """


class _WorkerCommunicationError(Exception):
    """An exception raised by communicators when a worker has died unexpectedly
    or become unresponsive.
    """


class _Sentinel(enum.Enum):
    """Sentinel values used to indicate sequence points or worker shutdown
    conditions.
    """

    NO_MORE_SCAN_REQUESTS = enum.auto()
    """Sentinel sent from the supervisor to scanners to indicate that there are
    no more quanta left to be scanned.
    """

    NO_MORE_INGEST_REQUESTS = enum.auto()
    """Sentinel sent from scanners to the ingester to indicate that there are
    will be no more ingest requests from a particular worker.
    """

    NO_MORE_WRITE_REQUESTS = enum.auto()
    """Sentinel sent from scanners and the supervisor to the writer to
    indicate that there are will be no more write requests from a particular
    worker.
    """

    WRITE_REPORT = enum.auto()
    """Sentinel sent from the writer to the supervisor to report that a
    quantum's provenance was written.
    """


@dataclasses.dataclass
class _WorkerErrorMessage:
    """An internal worker used to pass information about an error that occurred
    on a worker back to the supervisor.

    As a rule, these are unexpected, unrecoverable exceptions.
    """

    worker: str
    """Name of the originating worker."""

    traceback: str
    """A logged exception traceback.

    Note that this is not a `BaseException` subclass that can actually be
    re-raised on the supervisor; it's just something we can log to make the
    right traceback appear on the screen.  If something silences that printing
    in favor of its own exception management (pytest!) this information
    disappears.
    """


@dataclasses.dataclass
class _ScanRequest:
    """An internal struct passed from the supervisor to the scanners to request
    a quantum be scanned.
    """

    quantum_id: uuid.UUID
    """ID of the quantum to be scanned."""


@dataclasses.dataclass
class _IngestReport:
    """An internal struct passed from the ingester to the supervisor to report
    a completed ingest batch.
    """

    n_producers: int
    """Number of producing quanta whose datasets were ingested.

    We use quanta rather than datasets as the count here because the supervisor
    knows the total number of quanta in advance but not the total number of
    datasets to be ingested, so it's a lot easier to attach a denominator
    and/or progress bar to this number.
    """


@dataclasses.dataclass
class _WorkerDone:
    """An internal struct passed from a worker to the supervisor when it has
    successfully completed all work.
    """

    name: str
    """Name of the worker reporting completion."""


@dataclasses.dataclass
class _ProgressLog:
    """A high-level log message sent from a worker to the supervisor.

    These are messages that should appear to come from the main
    'aggregate-graph' logger, not a worker-specific one.
    """

    message: str
    """Log message."""

    level: int
    """Log level."""


@dataclasses.dataclass
class _CompressionDictionary:
    """An internal struct used to send the compression dictionary from the
    writer to the scanners.
    """

    data: bytes
    """The `bytes` representation of a `zstandard.ZstdCompressionDict`.
    """


type Report = (
    ScanReport
    | _IngestReport
    | _WorkerErrorMessage
    | _ProgressLog
    | _WorkerDone
    | Literal[_Sentinel.WRITE_REPORT]
)


def _disable_resources_parallelism() -> None:
    os.environ["LSST_RESOURCES_NUM_WORKERS"] = "1"
    os.environ.pop("LSST_RESOURCES_EXECUTOR", None)
    os.environ["LSST_S3_USE_THREADS"] = "False"


class SupervisorCommunicator:
    """A helper object that lets the supervisor direct the other workers.

    Parameters
    ----------
    log : `lsst.utils.logging.LsstLogAdapter`
        LSST-customized logger.
    n_scanners : `int`
        Number of scanner workers.
    worker_factory : `WorkerFactory`
        Abstraction over threading vs. multiprocessing.
    config : `AggregatorConfig`
        Configuration for the aggregator.
    """

    def __init__(
        self,
        log: LsstLogAdapter,
        n_scanners: int,
        worker_factory: WorkerFactory,
        config: AggregatorConfig,
    ) -> None:
        self.config = config
        self.progress = ProgressManager(log, config)
        self.n_scanners = n_scanners
        # The supervisor sends scan requests to scanners on this queue.
        # When complete, the supervisor sends n_scanners sentinals and each
        # scanner is careful to only take one before it starts its shutdown.
        self._scan_requests: Queue[_ScanRequest | Literal[_Sentinel.NO_MORE_SCAN_REQUESTS]] = (
            worker_factory.make_queue()
        )
        # The scanners send ingest requests to the ingester on this queue. Each
        # scanner sends one sentinal when it is done, and the ingester is
        # careful to wait for n_scanners sentinals to arrive before it starts
        # its shutdown.
        self._ingest_requests: Queue[IngestRequest | Literal[_Sentinel.NO_MORE_INGEST_REQUESTS]] = (
            worker_factory.make_queue()
        )
        # The scanners send write requests to the writer on this queue (which
        # will be `None` if we're not writing).  The supervisor also sends
        # write requests for blocked quanta (which we don't scan).  Each
        # scanner and the supervisor send one sentinal when done, and the
        # writer waits for (n_scanners + 1) sentinals to arrive before it
        # starts its shutdown.
        self._write_requests: (
            Queue[ProvenanceQuantumScanData | Literal[_Sentinel.NO_MORE_WRITE_REQUESTS]] | None
        ) = worker_factory.make_queue() if config.is_writing_provenance else None
        # All other workers use this queue to send many different kinds of
        # reports the supervisor.  The supervisor waits for a _DONE sentinal
        # from each worker before it finishes its shutdown.
        self._reports: Queue[Report] = worker_factory.make_queue()
        # The writer sends the compression dictionary to the scanners on this
        # queue.  It puts n_scanners copies on the queue, and each scanner only
        # takes one.  The compression_dict queue has no sentinal because it is
        # only used at most once; the supervisor takes responsibility for
        # clearing it out shutting down.
        self._compression_dict: Queue[_CompressionDictionary] = worker_factory.make_queue()
        # The supervisor sets this event when it receives an interrupt request
        # from an exception in the main process (usually KeyboardInterrupt).
        # Worker communicators check this in their polling loops and raise
        # FatalWorkerError when they see it set.
        self._cancel_event: Event = worker_factory.make_event()
        # Track what state we are in closing down, so we can start at the right
        # point if we're interrupted and __exit__ needs to clean up.  Note that
        # we can't rely on a non-exception __exit__ to do any shutdown work
        # that might be slow, since a KeyboardInterrupt that occurs when
        # __exit__ is already running can't be caught inside __exit__.
        self._sent_no_more_scan_requests = False
        self._sent_no_more_write_requests = False
        self._n_scanners_done = 0
        self.workers: dict[str, Worker] = {}

    def _wait_for_workers_to_finish(self, already_failing: bool = False) -> None:
        # Orderly shutdown, including exceptions: let workers clear out the
        # queues they're responsible for reading from.
        if not self._sent_no_more_scan_requests:
            for _ in range(self.n_scanners):
                self._scan_requests.put(_Sentinel.NO_MORE_SCAN_REQUESTS)
            self._sent_no_more_scan_requests = True
        if not self._sent_no_more_write_requests and self._write_requests is not None:
            self._write_requests.put(_Sentinel.NO_MORE_WRITE_REQUESTS)
            self._sent_no_more_write_requests = True
        while not all(w.successful for w in self.workers.values()):
            match self._handle_progress_reports(
                self._get_report(block=True), already_failing=already_failing
            ):
                case None | ScanReport():
                    pass
                case _WorkerDone(name=worker_name):
                    self.workers[worker_name].successful = True
                    if worker_name == IngesterCommunicator.get_worker_name():
                        self.progress.quantum_ingests.close()
                    elif worker_name == WriterCommunicator.get_worker_name():
                        self.progress.writes.close()
                    else:
                        self._n_scanners_done += 1
                        if self._n_scanners_done == self.n_scanners:
                            self.progress.scans.close()
                case unexpected:
                    raise AssertionError(f"Unexpected message {unexpected!r} to supervisor.")
            self.log.verbose(
                "Waiting for workers [%s] to report successful completion.",
                ", ".join(w.name for w in self.workers.values() if not w.successful),
            )
        self.log.verbose("Checking that all queues are empty.")
        if self._scan_requests.clear():
            self.progress.log.warning("Scan request queue was not empty at shutdown.")
            self._scan_requests.kill()
        if self._ingest_requests.clear():
            self.progress.log.warning("Ingest request queue was not empty at shutdown.")
            self._ingest_requests.kill()
        if self._write_requests is not None and self._write_requests.clear():
            self.progress.log.warning("Write request queue was not empty at shutdown.")
            self._write_requests.kill()
        if self._reports.clear():
            self.progress.log.warning("Reports queue was not empty at shutdown.")
            self._reports.kill()
        if self._compression_dict.clear():
            self.progress.log.warning("Compression dictionary queue was not empty at shutdown.")
            self._compression_dict.kill()
        for worker in self.workers.values():
            self.log.verbose("Waiting for %s to shut down.", worker.name)
            worker.join()

    def _terminate(self) -> None:
        # Disorderly shutdown: we cannot assume any of the
        # multiprocessing.Queue object work, and in fact they may hang
        # if we try to do anything with them.
        self._scan_requests.kill()
        self._ingest_requests.kill()
        if self._write_requests is not None:
            self._write_requests.kill()
        self._compression_dict.kill()
        self._reports.kill()
        for name, worker in self.workers.items():
            if worker.is_alive():
                self.progress.log.critical("Terminating worker %r.", name)
                worker.kill()

    def __enter__(self) -> Self:
        _disable_resources_parallelism()
        self.progress.__enter__()
        # We make the low-level logger in __enter__ instead of __init__ only
        # because that's the pattern used by true workers (where it matters).
        self.log = make_worker_log("supervisor", self.config)
        return self

    def __exit__(
        self,
        exc_type: type[BaseException] | None,
        exc_value: BaseException | None,
        traceback: TracebackType | None,
    ) -> None:
        if exc_type is not None:
            self._cancel_event.set()
            if exc_type is _WorkerCommunicationError:
                self.progress.log.critical("Worker '%s' was terminated before it could finish.", exc_value)
                self._terminate()
                return None
            if exc_type is not FatalWorkerError:
                self.progress.log.critical("Caught %s; attempting to shut down cleanly.", exc_type)
        try:
            self._wait_for_workers_to_finish(already_failing=exc_type is not None)
        except _WorkerCommunicationError as err:
            self.progress.log.critical(
                "Worker '%s' was terminated before it could finish (after scanning).", err
            )
            self._terminate()
            raise
        self.progress.__exit__(exc_type, exc_value, traceback)
        return None

    def request_scan(self, quantum_id: uuid.UUID) -> None:
        """Send a request to the scanners to scan the given quantum.

        Parameters
        ----------
        quantum_id : `uuid.UUID`
            ID of the quantum to scan.
        """
        self._scan_requests.put(_ScanRequest(quantum_id))

    def request_write(self, request: ProvenanceQuantumScanData) -> None:
        """Send a request to the writer to write provenance for the given scan.

        Parameters
        ----------
        request : `ProvenanceQuantumScanData`
            Information from scanning a quantum (or knowing you don't have to,
            in the case of blocked quanta).
        """
        assert self._write_requests is not None, "Writer should not be used if writing is disabled."
        self._write_requests.put(request)

    def poll(self) -> Iterator[ScanReport]:
        """Poll for reports from workers while sending scan requests.

        Yields
        ------
        scan_report : `ScanReport`
            A report from a scanner that a quantum was scanned.

        Notes
        -----
        This iterator blocks until the first scan report is received, and then
        it continues until the report queue is empty.
        """
        block = True
        while report := self._get_report(block=block):
            match self._handle_progress_reports(report):
                case ScanReport() as scan_report:
                    block = False
                    yield scan_report
                case None:
                    pass
                case unexpected:
                    raise AssertionError(f"Unexpected message {unexpected!r} to supervisor.")

    @overload
    def _get_report(self, block: Literal[True]) -> Report: ...

    @overload
    def _get_report(self, block: bool) -> Report | None: ...

    def _get_report(self, block: bool) -> Report | None:
        """Get a report from the reports queue, with timeout guards on
        blocking requests.

        This method may *return* WorkerCommunicatorError (rather than raise it)
        when a serious error occurred communicating with a subprocess.  This
        is to avoid raising an exception in an __exit__ method (which calls
        method).
        """
        report = self._reports.get(block=block, timeout=self.config.worker_check_timeout)
        while report is None and block:
            # We hit the timeout; make sure all of the workers
            # that should be alive actually are.
            for name, worker in self.workers.items():
                if not worker.successful and not worker.is_alive():
                    # Delete this worker from the list of workers so we don't
                    # hit this condition again when we try to handle the
                    # exception we raise.
                    raise _WorkerCommunicationError(name)
            # If nothing is dead and we didn't hit the hang timeout, keep
            # trying.
            report = self._reports.get(block=block, timeout=self.config.worker_check_timeout)
        return report

    def _handle_progress_reports(
        self, report: Report, already_failing: bool = False
    ) -> ScanReport | _WorkerDone | None:
        """Handle reports to the supervisor that can appear at any time, and
        are typically just updates to the progress we've made.

        This includes:

        - exceptions from workers (which raise `FatalWorkerError` here to
          trigger ``__exit__``);
        - ingest reports;
        - write reports;
        - progress logs.

        If one of these is handled, `None` is returned; otherwise the original
        report is returned.
        """
        match report:
            case _WorkerErrorMessage(traceback=traceback, worker=worker):
                self.progress.log.fatal("Exception raised on %s: \n%s", worker, traceback)
                if not already_failing:
                    raise FatalWorkerError()
            case _IngestReport(n_producers=n_producers):
                self.progress.quantum_ingests.update(n_producers)
            case _Sentinel.WRITE_REPORT:
                self.progress.writes.update(1)
            case _ProgressLog(message=message, level=level):
                self.progress.log.log(level, "%s [after %0.1fs]", message, self.progress.elapsed_time)
            case _:
                return report
        return None


class WorkerCommunicator:
    """A base class for non-supervisor worker communicators.

    Parameters
    ----------
    supervisor : `SupervisorCommunicator`
        Communicator for the supervisor to grab queues and information from.
    name : `str`
        Human-readable name for this worker.

    Notes
    -----
    Each worker communicator is constructed in the main process and entered as
    a context manager *only* on the actual worker process, so attributes that
    cannot be pickled are constructed in ``__enter__`` instead of ``__init__``.

    Worker communicators provide access to an `AggregatorConfig` and a logger
    to their workers.  As context managers, they handle exceptions and ensure
    clean shutdowns, and since most workers need to use a lot of other context
    managers (for file reading and writing, mostly), they provide an `enter`
    method to keep every worker from also having to be a context manager just
    to hold a context manager instance attribute.

    Worker communicators can also be configured to record and dump profiling
    information.
    """

    def __init__(self, supervisor: SupervisorCommunicator, name: str):
        self.name = name
        self.config = supervisor.config
        self._reports = supervisor._reports
        self._cancel_event = supervisor._cancel_event

    def __enter__(self) -> Self:
        _disable_resources_parallelism()
        self.log = make_worker_log(self.name, self.config)
        self.log.verbose("%s has PID %s (parent is %s).", self.name, os.getpid(), os.getppid())
        self._exit_stack = ExitStack().__enter__()
        if self.config.n_processes > 1:
            # Multiprocessing: ignore interrupts so we can shut down cleanly.
            signal.signal(signal.SIGINT, signal.SIG_IGN)
            if self.config.worker_profile_dir is not None:
                # We use time.time because we're interested in wall-clock time,
                # not just CPU effort, since this is I/O-bound work.
                self._profiler = cProfile.Profile(timer=time.time)
                self._profiler.enable()
        return self

    def __exit__(
        self,
        exc_type: type[BaseException] | None,
        exc_value: BaseException | None,
        traceback: TracebackType | None,
    ) -> bool | None:
        if self.config.worker_profile_dir is not None and self.config.n_processes > 1:
            self._profiler.disable()
            os.makedirs(self.config.worker_profile_dir, exist_ok=True)
            self._profiler.dump_stats(os.path.join(self.config.worker_profile_dir, f"{self.name}.profile"))
        if exc_value is not None:
            assert exc_type is not None, "Should be guaranteed by Python, but MyPy doesn't know that."
            if exc_type is not FatalWorkerError:
                self.log.warning("Error raised on this worker.", exc_info=(exc_type, exc_value, traceback))
                assert exc_type is not None and traceback is not None
                self._reports.put(
                    _WorkerErrorMessage(
                        self.name,
                        "".join(format_exception(exc_type, exc_value, traceback)),
                    )
                )
                self.log.debug("Error message sent to supervisor.")
            else:
                self.log.warning("Shutting down due to exception raised on another worker.")
        self._exit_stack.__exit__(exc_type, exc_value, traceback)
        return True

    @property
    def exit_stack(self) -> ExitStack:
        """A `contextlib.ExitStack` tied to the communicator."""
        return self._exit_stack

    def log_progress(self, level: int, message: str) -> None:
        """Send a high-level log message to the supervisor.

        Parameters
        ----------
        level : `int`
            Log level.  Should be ``VERBOSE`` or higher.
        message : `str`
            Log message.
        """
        self._reports.put(_ProgressLog(message=message, level=level))

    def check_for_cancel(self) -> None:
        """Check for a cancel signal from the supervisor and raise
        `FatalWorkerError` if it is present.
        """
        if self._cancel_event.is_set():
            raise FatalWorkerError()


class ScannerCommunicator(WorkerCommunicator):
    """A communicator for scanner workers.

    Parameters
    ----------
    supervisor : `SupervisorCommunicator`
        Communicator for the supervisor to grab queues and information from.
    scanner_id : `int`
        Integer ID for this canner.
    """

    def __init__(self, supervisor: SupervisorCommunicator, scanner_id: int):
        super().__init__(supervisor, self.get_worker_name(scanner_id))
        self.scanner_id = scanner_id
        self._scan_requests = supervisor._scan_requests
        self._ingest_requests = supervisor._ingest_requests
        self._write_requests = supervisor._write_requests
        self._compression_dict = supervisor._compression_dict
        self._got_no_more_scan_requests: bool = False
        self._sent_no_more_ingest_requests: bool = False

    @staticmethod
    def get_worker_name(scanner_id: int) -> str:
        return f"scanner-{scanner_id:03d}"

    def report_scan(self, msg: ScanReport) -> None:
        """Report a completed scan to the supervisor.

        Parameters
        ----------
        msg : `ScanReport`
            Report to send.
        """
        self._reports.put(msg)

    def request_ingest(self, request: IngestRequest) -> None:
        """Ask the ingester to ingest a quantum's outputs.

        Parameters
        ----------
        request : `IngestRequest`
            Description of the datasets to ingest.

        Notes
        -----
        If this request has no datasets, this automatically reports the ingest
        as complete to the supervisor instead of sending it to the ingester.
        """
        if request:
            self._ingest_requests.put(request)
        else:
            self._reports.put(_IngestReport(1))

    def request_write(self, request: ProvenanceQuantumScanData) -> None:
        """Ask the writer to write provenance for a quantum.

        Parameters
        ----------
        request : `ProvenanceQuantumScanData`
            Result of scanning a quantum.
        """
        assert self._write_requests is not None, "Writer should not be used if writing is disabled."
        self._write_requests.put(request)

    def get_compression_dict(self) -> bytes | None:
        """Attempt to get the compression dict from the writer.

        Returns
        -------
        data : `bytes` or `None`
            The `bytes` representation of the compression dictionary, or `None`
            if the compression dictionary is not yet available.

        Notes
        -----
        A scanner should only call this method before it actually has the
        compression dict.
        """
        if (cdict := self._compression_dict.get()) is not None:
            return cdict.data
        return None

    def poll(self) -> Iterator[uuid.UUID]:
        """Poll for scan requests to process.

        Yields
        ------
        quantum_id : `uuid.UUID`
            ID of a new quantum to scan.

        Notes
        -----
        This iterator ends when the supervisor reports that it is done
        traversing the graph.
        """
        while True:
            self.check_for_cancel()
            scan_request = self._scan_requests.get(block=True, timeout=self.config.worker_sleep)
            if scan_request is _Sentinel.NO_MORE_SCAN_REQUESTS:
                self._got_no_more_scan_requests = True
                return
            if scan_request is not None:
                yield scan_request.quantum_id

    def __exit__(
        self,
        exc_type: type[BaseException] | None,
        exc_value: BaseException | None,
        traceback: TracebackType | None,
    ) -> bool | None:
        result = super().__exit__(exc_type, exc_value, traceback)
        self._ingest_requests.put(_Sentinel.NO_MORE_INGEST_REQUESTS)
        if self._write_requests is not None:
            self._write_requests.put(_Sentinel.NO_MORE_WRITE_REQUESTS)
        while not self._got_no_more_scan_requests:
            if (
                not self._got_no_more_scan_requests
                and self._scan_requests.get(block=True) is _Sentinel.NO_MORE_SCAN_REQUESTS
            ):
                self._got_no_more_scan_requests = True
        # We let the writer clear out the compression dict queue.
        self.log.verbose("Sending completion message.")
        self._reports.put(_WorkerDone(self.name))
        return result


class IngesterCommunicator(WorkerCommunicator):
    """A communicator for the ingester worker.

    Parameters
    ----------
    supervisor : `SupervisorCommunicator`
        Communicator for the supervisor to grab queues and information from.
    """

    def __init__(self, supervisor: SupervisorCommunicator):
        super().__init__(supervisor, self.get_worker_name())
        self.n_scanners = supervisor.n_scanners
        self._ingest_requests = supervisor._ingest_requests
        self._n_requesters_done = 0

    @staticmethod
    def get_worker_name() -> str:
        return "ingester"

    def __exit__(
        self,
        exc_type: type[BaseException] | None,
        exc_value: BaseException | None,
        traceback: TracebackType | None,
    ) -> bool | None:
        result = super().__exit__(exc_type, exc_value, traceback)
        while self._n_requesters_done != self.n_scanners:
            self.log.debug(
                "Waiting for %d requesters to be done (currently %d).",
                self.n_scanners,
                self._n_requesters_done,
            )
            if self._ingest_requests.get(block=True) is _Sentinel.NO_MORE_INGEST_REQUESTS:
                self._n_requesters_done += 1
        self.log.verbose("Sending completion message.")
        self._reports.put(_WorkerDone(self.name))
        return result

    def report_ingest(self, n_producers: int) -> None:
        """Report to the supervisor that an ingest batch was completed.

        Parameters
        ----------
        n_producers : `int`
            Number of producing quanta whose datasets were ingested.
        """
        self._reports.put(_IngestReport(n_producers))

    def poll(self) -> Iterator[IngestRequest]:
        """Poll for ingest requests from the scanner workers.

        Yields
        ------
        request : `IngestRequest`
            A request to ingest datasets produced by a single quantum.

        Notes
        -----
        This iterator ends when all scanners indicate that they are done making
        ingest requests.
        """
        while True:
            self.check_for_cancel()
            ingest_request = self._ingest_requests.get(block=True, timeout=_TINY_TIMEOUT)
            if ingest_request is _Sentinel.NO_MORE_INGEST_REQUESTS:
                self._n_requesters_done += 1
                if self._n_requesters_done == self.n_scanners:
                    return
                else:
                    continue
            if ingest_request is not None:
                yield ingest_request


class WriterCommunicator(WorkerCommunicator):
    """A communicator for the writer worker.

    Parameters
    ----------
    supervisor : `SupervisorCommunicator`
        Communicator for the supervisor to grab queues and information from.
    """

    def __init__(self, supervisor: SupervisorCommunicator):
        assert supervisor._write_requests is not None
        super().__init__(supervisor, self.get_worker_name())
        self.n_scanners = supervisor.n_scanners
        self._write_requests = supervisor._write_requests
        self._compression_dict = supervisor._compression_dict
        self._n_requesters = supervisor.n_scanners + 1
        self._n_requesters_done = 0
        self._sent_compression_dict = False

    @staticmethod
    def get_worker_name() -> str:
        return "writer"

    def __exit__(
        self,
        exc_type: type[BaseException] | None,
        exc_value: BaseException | None,
        traceback: TracebackType | None,
    ) -> bool | None:
        result = super().__exit__(exc_type, exc_value, traceback)
        if exc_type is None:
            self.log_progress(logging.INFO, "Provenance quantum graph written successfully.")
        while self._n_requesters_done != self._n_requesters:
            self.log.debug(
                "Waiting for %d requesters to be done (currently %d).",
                self._n_requesters,
                self._n_requesters_done,
            )
            if self._write_requests.get(block=True) is _Sentinel.NO_MORE_WRITE_REQUESTS:
                self._n_requesters_done += 1
        if self._compression_dict.clear():
            self.log.verbose("Cleared out compression dictionary queue.")
        else:
            self.log.verbose("Compression dictionary queue was already empty.")
        self.log.verbose("Sending completion message.")
        self._reports.put(_WorkerDone(self.name))
        return result

    def poll(self) -> Iterator[ProvenanceQuantumScanData]:
        """Poll for writer requests from the scanner workers and supervisor.

        Yields
        ------
        request : `ProvenanceQuantumScanData`
            The result of a quantum scan.

        Notes
        -----
        This iterator ends when all scanners and the supervisor indicate that
        they are done making write requests.
        """
        while True:
            self.check_for_cancel()
            write_request = self._write_requests.get(block=True, timeout=_TINY_TIMEOUT)
            if write_request is _Sentinel.NO_MORE_WRITE_REQUESTS:
                self._n_requesters_done += 1
                if self._n_requesters_done == self._n_requesters:
                    return
                else:
                    continue
            if write_request is not None:
                yield write_request

    def send_compression_dict(self, cdict_data: bytes) -> None:
        """Send the compression dictionary to the scanners.

        Parameters
        ----------
        cdict_data : `bytes`
            The `bytes` representation of the compression dictionary.
        """
        self.log.debug("Sending compression dictionary.")
        for _ in range(self.n_scanners):
            self._compression_dict.put(_CompressionDictionary(cdict_data))
        self._sent_compression_dict = True

    def report_write(self) -> None:
        """Report to the supervisor that provenance for a quantum was written
        to the graph.
        """
        self._reports.put(_Sentinel.WRITE_REPORT)

    def periodically_check_for_cancel[T](self, iterable: Iterable[T], n: int = 100) -> Iterator[T]:
        """Iterate while checking for a cancellation signal every ``n``
        iterations.

        Parameters
        ----------
        iterable : `~collections.abc.Iterable`
            Object to iterate over.
        n : `int`
            Check for cancellation every ``n`` iterations.

        Returns
        -------
        iterator : `~collections.abc.Iterator`
            Iterator.
        """
        i = 0
        for entry in iterable:
            yield entry
            i += 1
            if i % n == 0:
                self.check_for_cancel()



# This file is part of pipe_base.
#
# Developed for the LSST Data Management System.
# This product includes software developed by the LSST Project
# (http://www.lsst.org).
# See the COPYRIGHT file at the top-level directory of this distribution
# for details of code ownership.
#
# This software is dual licensed under the GNU General Public License and also
# under a 3-clause BSD license. Recipients may choose which of these licenses
# to use; please see the files gpl-3.0.txt and/or bsd_license.txt,
# respectively.  If you choose the GPL option then the following text applies
# (but note that there is still no warranty even if you opt for BSD instead):
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

__all__ = ["ExecutionGraphFixup"]

import uuid
from abc import ABC
from collections.abc import Mapping

import networkx

from lsst.daf.butler import DataCoordinate

from .graph import QuantumGraph


class ExecutionGraphFixup(ABC):
    """Interface for classes which update quantum graphs before execution.

    Notes
    -----
    The primary goal of this class is to modify quanta dependencies which may
    not be possible to reflect in a quantum graph using standard tools. One
    known use case for that is to guarantee particular execution order of
    visits in CI jobs for cases when outcome depends on the processing order of
    visits (e.g. AP association pipeline).

    Instances of this class receive a preliminary graph and are allowed to
    add edges, as long as those edges do not result in a cycle.  Edges and
    nodes may not be removed.

    New subclasses should implement only `fixup_graph`, which will always be
    called first.  `fixupQuanta` is only called if `fixup_graph` raises
    `NotImplementedError`.
    """

    def fixupQuanta(self, graph: QuantumGraph) -> QuantumGraph:
        """Update quanta in a graph.

        Parameters
        ----------
        graph : `.QuantumGraph`
            Quantum Graph that will be executed by the executor.

        Returns
        -------
        graph : `.QuantumGraph`
            Modified graph.

        Notes
        -----
        This hook is provided for backwards compatibility only.
        """
        raise NotImplementedError()

    def fixup_graph(
        self, xgraph: networkx.DiGraph, quanta_by_task: Mapping[str, Mapping[DataCoordinate, uuid.UUID]]
    ) -> None:
        """Update a networkx graph of quanta in place by adding edges to
        further constrain the ordering.

        Parameters
        ----------
        xgraph : `networkx.DiGraph`
            A directed acyclic graph of quanta to modify in place.  Node keys
            are quantum UUIDs, and attributes include ``task_label`` (`str`)
            and ``data_id`` (a full `lsst.daf.butler.DataCoordinate`, without
            dimension records attached).  Edges may be added, but not removed.
            Nodes may not be modified.
        quanta_by_task : `~collections.abc.Mapping` [ `str`,\
                `~collections.abc.Mapping` [ `lsst.daf.butler.DataCoordinate`,\
                `uuid.UUID` ] ]
            All quanta in the graph, grouped first by task label and then by
            data ID.
        """
        raise NotImplementedError()



# This file is part of pipe_base.
#
# Developed for the LSST Data Management System.
# This product includes software developed by the LSST Project
# (https://www.lsst.org).
# See the COPYRIGHT file at the top-level directory of this distribution
# for details of code ownership.
#
# This software is dual licensed under the GNU General Public License and also
# under a 3-clause BSD license. Recipients may choose which of these licenses
# to use; please see the files gpl-3.0.txt and/or bsd_license.txt,
# respectively.  If you choose the GPL option then the following text applies
# (but note that there is still no warranty even if you opt for BSD instead):
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <https://www.gnu.org/licenses/>.

from __future__ import annotations

import logging
import multiprocessing
import multiprocessing.context
import os
import signal
import sys
import unittest
import warnings
from typing import Literal

import psutil

from lsst.pipe.base.exec_fixup_data_id import ExecFixupDataId
from lsst.pipe.base.mp_graph_executor import MPGraphExecutor, MPGraphExecutorError, MPTimeoutError
from lsst.pipe.base.quantum_reports import ExecutionStatus, Report
from lsst.pipe.base.tests.mocks import (
    DynamicConnectionConfig,
    DynamicTestPipelineTask,
    DynamicTestPipelineTaskConfig,
    InMemoryRepo,
)

logging.basicConfig(level=logging.DEBUG)

_LOG = logging.getLogger(__name__)

TESTDIR = os.path.abspath(os.path.dirname(__file__))


class NoMultiprocessingTask(DynamicTestPipelineTask):
    """A test pipeline task that declares that it cannot be used in
    multiprocessing.
    """

    canMultiprocess = False


def _count_status(report: Report, status: ExecutionStatus) -> int:
    """Count number of quanta with a given status."""
    return len([qrep for qrep in report.quantaReports if qrep.status is status])


class MPGraphExecutorTestCase(unittest.TestCase):
    """A test case for MPGraphExecutor class."""

    def test_mpexec_nomp(self) -> None:
        """Make simple graph and execute."""
        helper = InMemoryRepo("base.yaml")
        self.enterContext(helper)
        helper.add_task(dimensions=["detector"])
        qgraph = helper.make_quantum_graph()
        qexec, butler = helper.make_single_quantum_executor()
        # run in single-process mode
        mpexec = MPGraphExecutor(num_proc=1, timeout=100, quantum_executor=qexec)
        mpexec.execute(qgraph)  # type: ignore[arg-type]
        self.assertCountEqual(
            [ref.dataId["detector"] for ref in butler.get_datasets("dataset_auto1")], [1, 2, 3, 4]
        )
        report = mpexec.getReport()
        assert report is not None
        self.assertEqual(report.status, ExecutionStatus.SUCCESS)
        self.assertIsNone(report.exitCode)
        self.assertIsNone(report.exceptionInfo)
        self.assertEqual(len(report.quantaReports), 4)
        self.assertTrue(all(qrep.status == ExecutionStatus.SUCCESS for qrep in report.quantaReports))
        self.assertTrue(all(qrep.exitCode is None for qrep in report.quantaReports))
        self.assertTrue(all(qrep.exceptionInfo is None for qrep in report.quantaReports))
        self.assertTrue(all(qrep.taskLabel == "task_auto1" for qrep in report.quantaReports))

    def test_mpexec_mp(self) -> None:
        """Make simple graph and execute."""
        helper = InMemoryRepo("base.yaml")
        self.enterContext(helper)
        helper.add_task(dimensions=["detector"])
        qg = helper.make_quantum_graph()
        qexec, butler = helper.make_single_quantum_executor()

        methods: list[Literal["spawn", "forkserver"]] = ["spawn"]
        if sys.platform == "linux":
            methods.append("forkserver")

        for method in methods:
            with self.subTest(startMethod=method):
                # Run in multi-process mode, the order of results is not
                # defined.
                mpexec = MPGraphExecutor(num_proc=3, timeout=100, quantum_executor=qexec, start_method=method)
                mpexec.execute(qg)  # type: ignore[arg-type]
                report = mpexec.getReport()
                assert report is not None
                self.assertEqual(report.status, ExecutionStatus.SUCCESS)
                self.assertIsNone(report.exitCode)
                self.assertIsNone(report.exceptionInfo)
                self.assertEqual(len(report.quantaReports), 4)
                self.assertTrue(all(qrep.status == ExecutionStatus.SUCCESS for qrep in report.quantaReports))
                self.assertTrue(all(qrep.exitCode == 0 for qrep in report.quantaReports))
                self.assertTrue(all(qrep.exceptionInfo is None for qrep in report.quantaReports))
                self.assertTrue(all(qrep.taskLabel == "task_auto1" for qrep in report.quantaReports))

    def test_mpexec_nompsupport(self) -> None:
        """Try to run MP for task that has no MP support which should fail."""
        helper = InMemoryRepo("base.yaml")
        self.enterContext(helper)
        helper.add_task(task_class=NoMultiprocessingTask, dimensions=["detector"])
        qg = helper.make_quantum_graph()
        qexec, butler = helper.make_single_quantum_executor()
        mpexec = MPGraphExecutor(num_proc=3, timeout=100, quantum_executor=qexec)
        with self.assertRaisesRegex(
            MPGraphExecutorError, "Task 'task_auto1' does not support multiprocessing"
        ):
            mpexec.execute(qg)  # type: ignore[arg-type]

    def test_mpexec_fixup(self) -> None:
        """Make simple graph and execute, add dependencies by executing fixup
        code.
        """
        helper = InMemoryRepo("base.yaml")
        self.enterContext(helper)
        helper.add_task(dimensions=["detector"])
        qg = helper.make_quantum_graph()
        for reverse in (False, True):
            qexec, butler = helper.make_single_quantum_executor()
            fixup = ExecFixupDataId("task_auto1", "detector", reverse=reverse)
            mpexec = MPGraphExecutor(
                num_proc=1, timeout=100, quantum_executor=qexec, execution_graph_fixup=fixup
            )
            mpexec.execute(qg)  # type: ignore[arg-type]
            expected = [1, 2, 3, 4]
            if reverse:
                expected = list(reversed(expected))
            self.assertEqual(
                [ref.dataId["detector"] for ref in butler.get_datasets("dataset_auto1")], expected
            )

    def test_mpexec_fixup_old_qg(self) -> None:
        """Test using an old QuantumGraph object to initialize the executor,
        with an ordering fixup.
        """
        helper = InMemoryRepo("base.yaml")
        self.enterContext(helper)
        helper.add_task(dimensions=["detector"])
        qgraph = helper.make_quantum_graph_builder().build(attach_datastore_records=False)
        for reverse in (False, True):
            qexec, butler = helper.make_single_quantum_executor()
            fixup = ExecFixupDataId("task_auto1", "detector", reverse=reverse)
            mpexec = MPGraphExecutor(
                num_proc=1, timeout=100, quantum_executor=qexec, execution_graph_fixup=fixup
            )
            mpexec.execute(qgraph)  # type: ignore[arg-type]
            expected = [1, 2, 3, 4]
            if reverse:
                expected = list(reversed(expected))
            self.assertEqual(
                [ref.dataId["detector"] for ref in butler.get_datasets("dataset_auto1")], expected
            )

    def test_mpexec_timeout(self) -> None:
        """Fail due to timeout."""
        helper = InMemoryRepo("base.yaml")
        self.enterContext(helper)
        helper.add_task(label="a")
        helper.add_task(
            label="b",
            inputs={"input_connection": DynamicConnectionConfig(dataset_type_name="dataset_auto0")},
        )
        helper.add_task(
            label="c",
            inputs={"input_connection": DynamicConnectionConfig(dataset_type_name="dataset_auto0")},
            config=DynamicTestPipelineTaskConfig(sleep=100.0),
        )
        qg = helper.make_quantum_graph()

        # with failFast we'll get immediate MPTimeoutError
        qexec, _ = helper.make_single_quantum_executor()
        mpexec = MPGraphExecutor(num_proc=3, timeout=1, quantum_executor=qexec, fail_fast=True)
        with self.assertRaises(MPTimeoutError):
            mpexec.execute(qg)  # type: ignore[arg-type]
        report = mpexec.getReport()
        assert report is not None and report.exceptionInfo is not None
        self.assertEqual(report.status, ExecutionStatus.TIMEOUT)
        self.assertEqual(report.exceptionInfo.className, "lsst.pipe.base.mp_graph_executor.MPTimeoutError")
        self.assertGreater(len(report.quantaReports), 0)
        self.assertEqual(_count_status(report, ExecutionStatus.TIMEOUT), 1)
        self.assertTrue(any(qrep.exitCode is not None and qrep.exitCode < 0 for qrep in report.quantaReports))
        self.assertTrue(all(qrep.exceptionInfo is None for qrep in report.quantaReports))

        # with failFast=False exception happens after last task finishes
        qexec, _ = helper.make_single_quantum_executor()
        mpexec = MPGraphExecutor(num_proc=3, timeout=3, quantum_executor=qexec, fail_fast=False)
        with self.assertRaises(MPTimeoutError):
            mpexec.execute(qg)  # type: ignore[arg-type]
        assert report is not None and report.exceptionInfo is not None
        self.assertEqual(report.status, ExecutionStatus.TIMEOUT)
        self.assertEqual(report.exceptionInfo.className, "lsst.pipe.base.mp_graph_executor.MPTimeoutError")
        # We expect two tasks ('a' and 'b') to finish successfully and one task
        # ('c') to timeout, which should get us all three reports.
        # Unfortunately on busy CPU there is no guarantee that tasks finish on
        # time, so expect more timeouts and issue a warning.
        if len(report.quantaReports) != 3:
            warnings.warn(
                f"Possibly timed out tasks, expected three reports, received {len(report.quantaReports)})."
            )
        report = mpexec.getReport()
        self.assertGreater(_count_status(report, ExecutionStatus.TIMEOUT), 0)
        self.assertTrue(any(qrep.exitCode is not None and qrep.exitCode < 0 for qrep in report.quantaReports))
        self.assertTrue(all(qrep.exceptionInfo is None for qrep in report.quantaReports))

    def test_mpexec_failure(self) -> None:
        """Failure in one task should not stop other tasks."""
        helper = InMemoryRepo("base.yaml")
        self.enterContext(helper)
        helper.add_task(
            config=DynamicTestPipelineTaskConfig(fail_condition="detector=2"),
            dimensions=["detector"],
        )
        qg = helper.make_quantum_graph()
        qexec, _ = helper.make_single_quantum_executor()
        mpexec = MPGraphExecutor(num_proc=3, timeout=100, quantum_executor=qexec)
        with self.assertRaisesRegex(MPGraphExecutorError, "One or more tasks failed"):
            mpexec.execute(qg)  # type: ignore[arg-type]
        report = mpexec.getReport()
        assert report is not None and report.exceptionInfo is not None
        self.assertEqual(report.status, ExecutionStatus.FAILURE)
        self.assertEqual(
            report.exceptionInfo.className, "lsst.pipe.base.mp_graph_executor.MPGraphExecutorError"
        )
        self.assertGreater(len(report.quantaReports), 0)
        self.assertEqual(_count_status(report, ExecutionStatus.FAILURE), 1)
        self.assertEqual(_count_status(report, ExecutionStatus.SUCCESS), 3)
        self.assertTrue(any(qrep.exitCode is not None and qrep.exitCode > 0 for qrep in report.quantaReports))
        self.assertTrue(any(qrep.exceptionInfo is not None for qrep in report.quantaReports))

    def test_mpexec_failure_dep(self) -> None:
        """Failure in one task should skip dependents."""
        helper = InMemoryRepo("base.yaml")
        self.enterContext(helper)
        helper.add_task(
            "a", config=DynamicTestPipelineTaskConfig(fail_condition="detector=2"), dimensions=["detector"]
        )
        helper.add_task("b", dimensions=["detector"])  # depends on 'a', for the same detector.
        qg = helper.make_quantum_graph()
        qexec, _ = helper.make_single_quantum_executor()
        mpexec = MPGraphExecutor(num_proc=3, timeout=100, quantum_executor=qexec)
        with self.assertRaisesRegex(MPGraphExecutorError, "One or more tasks failed"):
            mpexec.execute(qg)  # type: ignore[arg-type]
        report = mpexec.getReport()
        assert report is not None and report.exceptionInfo is not None
        self.assertEqual(report.status, ExecutionStatus.FAILURE)
        self.assertEqual(
            report.exceptionInfo.className, "lsst.pipe.base.mp_graph_executor.MPGraphExecutorError"
        )
        # Dependencies of failed tasks do not appear in quantaReports
        self.assertGreater(len(report.quantaReports), 0)
        self.assertEqual(_count_status(report, ExecutionStatus.FAILURE), 1)
        self.assertEqual(_count_status(report, ExecutionStatus.SUCCESS), 6)
        self.assertEqual(_count_status(report, ExecutionStatus.SKIPPED), 1)
        self.assertTrue(any(qrep.exitCode is not None and qrep.exitCode > 0 for qrep in report.quantaReports))
        self.assertTrue(any(qrep.exceptionInfo is not None for qrep in report.quantaReports))

    def test_mpexec_failure_dep_nomp(self) -> None:
        """Failure in one task should skip dependents, in-process version."""
        helper = InMemoryRepo("base.yaml")
        self.enterContext(helper)
        helper.add_task(
            "a", config=DynamicTestPipelineTaskConfig(fail_condition="detector=2"), dimensions=["detector"]
        )
        helper.add_task("b", dimensions=["detector"])  # depends on 'a', for the same detector.
        qg = helper.make_quantum_graph()
        qexec, butler = helper.make_single_quantum_executor()
        mpexec = MPGraphExecutor(num_proc=1, timeout=100, quantum_executor=qexec)
        with self.assertRaisesRegex(MPGraphExecutorError, "One or more tasks failed"):
            mpexec.execute(qg)  # type: ignore[arg-type]
        self.assertCountEqual(
            [ref.dataId["detector"] for ref in butler.get_datasets("dataset_auto1")], [1, 3, 4]
        )
        self.assertCountEqual(
            [ref.dataId["detector"] for ref in butler.get_datasets("dataset_auto2")], [1, 3, 4]
        )
        report = mpexec.getReport()
        assert report is not None and report.exceptionInfo is not None
        self.assertEqual(report.status, ExecutionStatus.FAILURE)
        self.assertEqual(
            report.exceptionInfo.className, "lsst.pipe.base.mp_graph_executor.MPGraphExecutorError"
        )
        # Dependencies of failed tasks do not appear in quantaReports
        self.assertGreater(len(report.quantaReports), 0)
        self.assertEqual(_count_status(report, ExecutionStatus.FAILURE), 1)
        self.assertEqual(_count_status(report, ExecutionStatus.SUCCESS), 6)
        self.assertEqual(_count_status(report, ExecutionStatus.SKIPPED), 1)
        self.assertTrue(all(qrep.exitCode is None for qrep in report.quantaReports))
        self.assertTrue(any(qrep.exceptionInfo is not None for qrep in report.quantaReports))

    def test_mpexec_failure_failfast(self) -> None:
        """Fast fail stops quickly.

        Timing delay of task 'b' should be sufficient to process
        failure and raise exception before task 'c'.
        """
        helper = InMemoryRepo("base.yaml")
        self.enterContext(helper)
        helper.add_task(
            "a", config=DynamicTestPipelineTaskConfig(fail_condition="detector=2"), dimensions=["detector"]
        )
        helper.add_task("b", config=DynamicTestPipelineTaskConfig(sleep=100.0), dimensions=["detector"])
        helper.add_task("c", dimensions=["detector"])  # depends on 'b', for the same detector.
        qg = helper.make_quantum_graph()
        qexec, _ = helper.make_single_quantum_executor()
        mpexec = MPGraphExecutor(num_proc=3, timeout=100, quantum_executor=qexec, fail_fast=True)
        with self.assertRaisesRegex(MPGraphExecutorError, "failed, exit code=1"):
            mpexec.execute(qg)  # type: ignore[arg-type]
        report = mpexec.getReport()
        assert report is not None and report.exceptionInfo is not None
        self.assertEqual(report.status, ExecutionStatus.FAILURE)
        self.assertEqual(
            report.exceptionInfo.className, "lsst.pipe.base.mp_graph_executor.MPGraphExecutorError"
        )
        # Dependencies of failed tasks do not appear in quantaReports
        self.assertGreater(len(report.quantaReports), 0)
        self.assertEqual(_count_status(report, ExecutionStatus.FAILURE), 1)
        self.assertTrue(any(qrep.exitCode is not None and qrep.exitCode > 0 for qrep in report.quantaReports))
        self.assertTrue(any(qrep.exceptionInfo is not None for qrep in report.quantaReports))

    def test_mpexec_crash(self) -> None:
        """Check task crash due to signal."""
        helper = InMemoryRepo("base.yaml")
        self.enterContext(helper)
        helper.add_task(
            config=DynamicTestPipelineTaskConfig(fail_condition="detector=2", fail_signal=signal.SIGILL),
            dimensions=["detector"],
        )
        qg = helper.make_quantum_graph()
        qexec, _ = helper.make_single_quantum_executor()
        mpexec = MPGraphExecutor(num_proc=3, timeout=100, quantum_executor=qexec)
        with self.assertRaisesRegex(MPGraphExecutorError, "One or more tasks failed"):
            mpexec.execute(qg)  # type: ignore[arg-type]
        report = mpexec.getReport()
        assert report is not None and report.exceptionInfo is not None
        self.assertEqual(report.status, ExecutionStatus.FAILURE)
        self.assertEqual(
            report.exceptionInfo.className, "lsst.pipe.base.mp_graph_executor.MPGraphExecutorError"
        )
        # Dependencies of failed tasks do not appear in quantaReports
        self.assertGreater(len(report.quantaReports), 0)
        self.assertEqual(_count_status(report, ExecutionStatus.FAILURE), 1)
        self.assertEqual(_count_status(report, ExecutionStatus.SUCCESS), 3)
        self.assertTrue(any(qrep.exitCode == -signal.SIGILL for qrep in report.quantaReports))
        self.assertTrue(all(qrep.exceptionInfo is None for qrep in report.quantaReports))

    def test_mpexec_crash_failfast(self) -> None:
        """Check task crash due to signal with --fail-fast."""
        helper = InMemoryRepo("base.yaml")
        self.enterContext(helper)
        helper.add_task(
            "a",
            config=DynamicTestPipelineTaskConfig(fail_condition="detector=2", fail_signal=signal.SIGILL),
            dimensions=["detector"],
        )
        helper.add_task("b", config=DynamicTestPipelineTaskConfig(sleep=100.0), dimensions=["detector"])
        helper.add_task("c", dimensions=["detector"])  # depends on 'b', for the same detector.
        qg = helper.make_quantum_graph()
        qexec, _ = helper.make_single_quantum_executor()
        mpexec = MPGraphExecutor(num_proc=3, timeout=100, quantum_executor=qexec, fail_fast=True)
        with self.assertRaisesRegex(MPGraphExecutorError, "failed, killed by signal 4 .Illegal instruction"):
            mpexec.execute(qg)  # type: ignore[arg-type]
        report = mpexec.getReport()
        assert report is not None and report.exceptionInfo is not None
        self.assertEqual(report.status, ExecutionStatus.FAILURE)
        self.assertEqual(
            report.exceptionInfo.className, "lsst.pipe.base.mp_graph_executor.MPGraphExecutorError"
        )
        self.assertEqual(_count_status(report, ExecutionStatus.FAILURE), 1)
        self.assertTrue(any(qrep.exitCode == -signal.SIGILL for qrep in report.quantaReports))
        self.assertTrue(all(qrep.exceptionInfo is None for qrep in report.quantaReports))

    def test_mpexec_num_fd(self) -> None:
        """Check that number of open files stays reasonable."""
        helper = InMemoryRepo("base.yaml")
        self.enterContext(helper)
        helper.add_task("a", task_class=NoMultiprocessingTask, dimensions=["detector", "visit"])
        helper.add_task("b", task_class=NoMultiprocessingTask, dimensions=["detector", "visit"])
        qg = helper.make_quantum_graph()
        qexec, _ = helper.make_single_quantum_executor()
        this_proc = psutil.Process()
        num_fds_0 = this_proc.num_fds()

        # run in multi-process mode, the order of results is not defined
        mpexec = MPGraphExecutor(num_proc=3, timeout=100, quantum_executor=qexec)
        mpexec.execute(qg)  # type: ignore[arg-type]

        num_fds_1 = this_proc.num_fds()
        # They should be the same but allow small growth just in case.
        # Without DM-26728 fix the difference would be equal to number of
        # quanta (20).
        self.assertLess(num_fds_1 - num_fds_0, 5)


def setup_module(module):
    """Force spawn to be used if no method given explicitly.

    This can be removed when Python 3.14 changes the default.

    Parameters
    ----------
    module : `~types.ModuleType`
        Module to set up.
    """
    multiprocessing.set_start_method("spawn", force=True)


if __name__ == "__main__":
    # Do not need to force start mode when running standalone.
    multiprocessing.set_start_method("spawn")
    unittest.main()



# This file is part of pipe_base.
#
# Developed for the LSST Data Management System.
# This product includes software developed by the LSST Project
# (http://www.lsst.org).
# See the COPYRIGHT file at the top-level directory of this distribution
# for details of code ownership.
#
# This software is dual licensed under the GNU General Public License and also
# under a 3-clause BSD license. Recipients may choose which of these licenses
# to use; please see the files gpl-3.0.txt and/or bsd_license.txt,
# respectively.  If you choose the GPL option then the following text applies
# (but note that there is still no warranty even if you opt for BSD instead):
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

__all__ = ["transfer_from_graph"]

import math

from lsst.daf.butler import Butler, CollectionType, MissingCollectionError, QuantumBackedButler
from lsst.pipe.base import QuantumGraph
from lsst.utils.iteration import chunk_iterable
from lsst.utils.logging import getLogger

from .utils import filter_by_dataset_type_glob, filter_by_existence

_LOG = getLogger(__name__)


def transfer_from_graph(
    graph: str,
    dest: str,
    register_dataset_types: bool,
    transfer_dimensions: bool,
    update_output_chain: bool,
    dry_run: bool,
    dataset_type: tuple[str, ...],
) -> int:
    """Transfer output datasets from quantum graph to dest.

    Parameters
    ----------
    graph : `str`
        URI string of the quantum graph.
    dest : `str`
        URI string of the destination Butler repo.
    register_dataset_types : `bool`
        Indicate whether missing dataset types should be registered.
    transfer_dimensions : `bool`
        Indicate whether dimensions should be transferred along with datasets.
        It can be more efficient to disable this if it is known that all
        dimensions exist.
    update_output_chain : `bool`
        If quantum graph metadata includes output run name and output
        collection which is a chain, update the chain definition to include run
        name as a the first collection in the chain.
    dry_run : `bool`
        Run the transfer without updating the destination butler.
    dataset_type : `tuple` of `str`
        Dataset type names. An empty tuple implies all dataset types.
        Can include globs.

    Returns
    -------
    count : `int`
        Actual count of transferred datasets.
    """
    # Read whole graph into memory
    qgraph = QuantumGraph.loadUri(graph)

    output_refs, _ = qgraph.get_refs(include_outputs=True, include_init_outputs=True, conform_outputs=True)

    # Get data repository dataset type definitions from the QuantumGraph.
    dataset_types = {dstype.name: dstype for dstype in qgraph.registryDatasetTypes()}

    # Filter the refs based on requested dataset types.
    filtered_refs = filter_by_dataset_type_glob(output_refs, dataset_type)
    _LOG.verbose("After filtering by dataset_type, number of datasets to transfer: %d", len(filtered_refs))

    # Make QBB, its config is the same as output Butler.
    with (
        QuantumBackedButler.from_predicted(
            config=dest,
            predicted_inputs=[ref.id for ref in output_refs],
            predicted_outputs=[],
            dimensions=qgraph.universe,
            datastore_records={},
            dataset_types=dataset_types,
        ) as qbb,
        Butler.from_config(dest, writeable=True) as dest_butler,
    ):
        # For faster restarts, filter out those the destination already knows.
        filtered_refs = filter_by_existence(dest_butler, filtered_refs)

        # Transfer in chunks
        chunk_size = 50_000
        n_chunks = math.ceil(len(filtered_refs) / chunk_size)
        chunk_num = 0
        count = 0
        for chunk in chunk_iterable(filtered_refs, chunk_size=chunk_size):
            chunk_num += 1
            if n_chunks > 1:
                _LOG.verbose("Transferring %d datasets in chunk %d/%d", len(chunk), chunk_num, n_chunks)
            transferred = dest_butler.transfer_from(
                qbb,
                chunk,
                transfer="auto",
                register_dataset_types=register_dataset_types,
                transfer_dimensions=transfer_dimensions,
                dry_run=dry_run,
            )
            count += len(transferred)

        # If asked to do so, update output chain definition.
        if update_output_chain and (metadata := qgraph.metadata) is not None:
            # These are defined in CmdLineFwk.
            output_run = metadata.get("output_run")
            output = metadata.get("output")
            input = metadata.get("input")
            if output_run is not None and output is not None:
                _update_chain(dest_butler, output, output_run, input)

    return count


def _update_chain(butler: Butler, output_chain: str, output_run: str, inputs: list[str] | None) -> None:
    """Update chain definition if it exists to include run as the first item
    in a chain. If it does not exist then create it to include all inputs and
    output.

    Parameters
    ----------
    butler : `lsst.daf.butler.Butler`
        Butler where to update the collection chain.
    output_chain : `str`
        Name of the output CHAINED collection.
    output_run : `str`
        Name of the output RUN collection.
    inputs : `list` [`str`] | None
        All the input collections to be included in the chain if the
        chain is created.
    """
    # Do not need to update chain if output_run does not already exist.
    try:
        _ = butler.collections.get_info(output_run)
    except MissingCollectionError:
        _LOG.verbose(
            "Output RUN collection (%s) does not exist.  Skipping updating the output chain.",
            output_run,
        )
        return

    # Make chain collection if doesn't exist before calling prepend_chain.
    created_now = butler.collections.register(output_chain, CollectionType.CHAINED)
    if created_now:
        _LOG.verbose("Registered chain collection: %s", output_chain)
        if inputs:
            # First must flatten any input chains
            flattened = butler.collections.query(inputs, flatten_chains=True)

            # Add input collections to chain collection just made.  Using
            # extend instead of prepend in case of race condition where another
            # execution adds a run before this adds the inputs to the chain.
            butler.collections.extend_chain(output_chain, flattened)
    _LOG.verbose(
        "Prepending output chain collection (%s) with output RUN collection (%s)", output_chain, output_run
    )
    butler.collections.prepend_chain(output_chain, output_run)



# This file is part of pipe_base.
#
# Developed for the LSST Data Management System.
# This product includes software developed by the LSST Project
# (http://www.lsst.org).
# See the COPYRIGHT file at the top-level directory of this distribution
# for details of code ownership.
#
# This software is dual licensed under the GNU General Public License and also
# under a 3-clause BSD license. Recipients may choose which of these licenses
# to use; please see the files gpl-3.0.txt and/or bsd_license.txt,
# respectively.  If you choose the GPL option then the following text applies
# (but note that there is still no warranty even if you opt for BSD instead):
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

"""Simple unit test for configOverrides."""

import copy
import pickle
import unittest

from lsst.daf.butler import DatasetRef, DatasetType, DimensionUniverse
from lsst.pipe.base import DeferredDatasetRef
from lsst.utils.tests import MemoryTestCase, init


class DeferredDatasetRefTestCase(unittest.TestCase):
    """A test case for  DeferredDatasetRef"""

    def test_checkCopy(self):
        dsr = DatasetRef(
            datasetType=DatasetType(
                "pvi", dimensions="visit", storageClass="dict", universe=DimensionUniverse()
            ),
            dataId=dict(visit=10, instrument="test"),
            run="na",
        )
        ddsr = DeferredDatasetRef(dsr)
        ddsr_copy = copy.deepcopy(ddsr)
        assert isinstance(ddsr_copy, DeferredDatasetRef)

        round_trip = pickle.loads(pickle.dumps(ddsr))
        assert isinstance(round_trip, DeferredDatasetRef)


class MyMemoryTestCase(MemoryTestCase):
    """Check for memory leaks."""


def setup_module(module):
    """Configure pytest."""
    init()


if __name__ == "__main__":
    init()
    unittest.main()



# This file is part of pipe_base.
#
# Developed for the LSST Data Management System.
# This product includes software developed by the LSST Project
# (http://www.lsst.org).
# See the COPYRIGHT file at the top-level directory of this distribution
# for details of code ownership.
#
# This software is dual licensed under the GNU General Public License and also
# under a 3-clause BSD license. Recipients may choose which of these licenses
# to use; please see the files gpl-3.0.txt and/or bsd_license.txt,
# respectively.  If you choose the GPL option then the following text applies
# (but note that there is still no warranty even if you opt for BSD instead):
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

from __future__ import annotations

__all__ = ("Event", "Queue", "SpawnWorkerFactory", "ThreadWorkerFactory", "Worker", "WorkerFactory")

import multiprocessing.context
import multiprocessing.synchronize
import queue
import threading
from abc import ABC, abstractmethod
from collections.abc import Callable
from typing import Any, Literal, overload

_TINY_TIMEOUT = 0.01

type Event = threading.Event | multiprocessing.synchronize.Event


class Worker(ABC):
    """A thin abstraction over `threading.Thread` and `multiprocessing.Process`
    that also provides a variable to track whether it reported successful
    completion.
    """

    def __init__(self) -> None:
        self.successful = False

    @property
    @abstractmethod
    def name(self) -> str:
        """Name of the worker, as assigned at creation."""
        raise NotImplementedError()

    @abstractmethod
    def join(self, timeout: float | None = None) -> None:
        """Wait for the worker to finish.

        Parameters
        ----------
        timeout : `float`, optional
            How long to wait in seconds.  If the timeout is exceeded,
            `is_alive` can be used to see whether the worker finished or not.
        """
        raise NotImplementedError()

    @abstractmethod
    def is_alive(self) -> bool:
        """Return whether the worker is still running."""
        raise NotImplementedError()

    def kill(self) -> None:
        """Kill the worker, if possible."""


class Queue[T](ABC):
    """A thin abstraction over `queue.Queue` and `multiprocessing.Queue` that
    provides better control over disorderly shutdowns.
    """

    @overload
    def get(self, *, block: Literal[True]) -> T: ...

    @overload
    def get(self, *, timeout: float | None = None, block: bool = False) -> T | None: ...

    @abstractmethod
    def get(self, *, timeout: float | None = None, block: bool = False) -> T | None:
        """Get an object or return `None` if the queue is empty.

        Parameters
        ----------
        timeout : `float` or `None`, optional
            Maximum number of seconds to wait while blocking.
        block : `bool`, optional
            Whether to block until an object is available.

        Returns
        -------
        obj : `object` or `None`
            Object from the queue, or `None` if it was empty.  Note that this
            is different from the behavior of the built-in Python queues,
            which raise `queue.Empty` instead.
        """
        raise NotImplementedError()

    @abstractmethod
    def put(self, item: T) -> None:
        """Add an object to the queue.

        Parameters
        ----------
        item : `object`
            Item to add.
        """
        raise NotImplementedError()

    def clear(self) -> bool:
        """Clear out all objects currently on the queue.

        This does not guarantee that more objects will not be added later.
        """
        found_anything: bool = False
        while self.get() is not None:
            found_anything = True
        return found_anything

    def kill(self) -> None:
        """Prepare a queue for a disorderly shutdown, without assuming that
        any other workers using it are still alive and functioning.
        """


class WorkerFactory(ABC):
    """A simple abstract interface that can be implemented by both threading
    and multiprocessing.
    """

    @abstractmethod
    def make_queue(self) -> Queue[Any]:
        """Make an empty queue that can be used to pass objects between
        workers created by this factory.
        """
        raise NotImplementedError()

    @abstractmethod
    def make_event(self) -> Event:
        """Make an event that can be used to communicate a boolean state change
        to workers created by this factory.
        """
        raise NotImplementedError()

    @abstractmethod
    def make_worker(
        self, target: Callable[..., None], args: tuple[Any, ...], name: str | None = None
    ) -> Worker:
        """Make a worker that runs the given callable.

        Parameters
        ----------
        target : `~collections.abc.Callable`
            A callable to invoke on the worker.
        args : `tuple`
            Positional arguments to pass to the callable.
        name : `str`, optional
            Human-readable name for the worker.

        Returns
        -------
        worker : `Worker`
            Process or thread that is already running the given callable.
        """
        raise NotImplementedError()


class _ThreadWorker(Worker):
    """An implementation of `Worker` backed by the `threading` module."""

    def __init__(self, thread: threading.Thread):
        super().__init__()
        self._thread = thread

    @property
    def name(self) -> str:
        return self._thread.name

    def join(self, timeout: float | None = None) -> None:
        self._thread.join(timeout=timeout)

    def is_alive(self) -> bool:
        return self._thread.is_alive()


class _ThreadQueue[T](Queue[T]):
    def __init__(self) -> None:
        self._impl = queue.Queue[T]()

    @overload
    def get(self, *, block: Literal[True]) -> T: ...

    @overload
    def get(self, *, timeout: float | None = None, block: bool = False) -> T | None: ...

    def get(self, *, timeout: float | None = None, block: bool = False) -> T | None:
        try:
            return self._impl.get(block=block, timeout=timeout)
        except queue.Empty:
            return None

    def put(self, item: T) -> None:
        self._impl.put(item, block=False)


class ThreadWorkerFactory(WorkerFactory):
    """An implementation of `WorkerFactory` backed by the `threading`
    module.
    """

    def make_queue(self) -> Queue[Any]:
        return _ThreadQueue()

    def make_event(self) -> Event:
        return threading.Event()

    def make_worker(
        self, target: Callable[..., None], args: tuple[Any, ...], name: str | None = None
    ) -> Worker:
        thread = threading.Thread(target=target, args=args, name=name)
        thread.start()
        return _ThreadWorker(thread)


class _ProcessWorker(Worker):
    """An implementation of `Worker` backed by the `multiprocessing` module."""

    def __init__(self, process: multiprocessing.context.SpawnProcess):
        super().__init__()
        self._process = process

    @property
    def name(self) -> str:
        return self._process.name

    def join(self, timeout: float | None = None) -> None:
        self._process.join(timeout=timeout)

    def is_alive(self) -> bool:
        return self._process.is_alive()

    def kill(self) -> None:
        """Kill the worker, if possible."""
        self._process.kill()


class _ProcessQueue[T](Queue[T]):
    def __init__(self, impl: multiprocessing.Queue):
        self._impl = impl

    @overload
    def get(self, *, block: Literal[True]) -> T: ...

    @overload
    def get(self, *, timeout: float | None = None, block: bool = False) -> T | None: ...

    def get(self, *, timeout: float | None = None, block: bool = False) -> T | None:
        try:
            return self._impl.get(block=block, timeout=timeout)
        except queue.Empty:
            return None

    def put(self, item: T) -> None:
        self._impl.put(item, block=False)

    def kill(self) -> None:
        self._impl.cancel_join_thread()
        self._impl.close()


class SpawnWorkerFactory(WorkerFactory):
    """An implementation of `WorkerFactory` backed by the `multiprocessing`
    module, with new processes started by spawning.
    """

    def __init__(self) -> None:
        self._ctx = multiprocessing.get_context("spawn")

    def make_queue(self) -> Queue[Any]:
        return _ProcessQueue(self._ctx.Queue())

    def make_event(self) -> Event:
        return self._ctx.Event()

    def make_worker(
        self, target: Callable[..., None], args: tuple[Any, ...], name: str | None = None
    ) -> Worker:
        process = self._ctx.Process(target=target, args=args, name=name)
        process.start()
        return _ProcessWorker(process)
